#include "Chiplet_Interface.h"
#include <iostream>

// Constructor implementation
ChipletInterface::ChipletInterface(sc_core::sc_module_name name, int num_pes): sc_module(name), num_pes(num_pes) {
    //sc_report_handler::set_log_file_name("out/report.log");
    sc_report_handler::set_actions(C_LOG, SC_INFO, SC_LOG|SC_DISPLAY);
    for (int i = 0; i < num_pes; ++i) {
        // Register b_transport method for each PE receive socket
        oec_receive_sockets[i].register_b_transport(this, &ChipletInterface::oec_b_transport);   
    }
    
    //SC_THREAD(send_psum_interposer);
    // Register b_transport method for Interposer interface
    interposer_features_receive_socket.register_b_transport(this, &ChipletInterface::ii_fetures_receive_b_transport);
    interposer_weight_receive_socket.register_b_transport(this, &ChipletInterface::ii_weight_receive_b_transport);
}

// Handle incoming transactions from the Interposer Interface
void ChipletInterface::ii_fetures_receive_b_transport(tlm_generic_payload& trans, sc_time& delay) {
    CustomPayload* orig_trans = dynamic_cast<CustomPayload*>(&trans);
    if (!orig_trans) {
        wait(5, SC_NS);
        log_error(std::to_string(chiplet_id) + " Invalid transaction type (not CustomPayload)." );
        trans.set_response_status(TLM_GENERIC_ERROR_RESPONSE);
        return;
    }

    CustomPayload* chip_trans1 = new CustomPayload();
    chip_trans1->update_timestamp(sc_time_stamp());
    sc_time current_time = sc_time_stamp();
    //manually copy the transaction payload for future use
    // Log the time when the core sent the transaction
    sc_time sent_time = orig_trans->timestamp;
    chip_trans1->data = orig_trans->data;
    chip_trans1->num_rows = orig_trans->num_rows;
    chip_trans1->num_cols = orig_trans->num_cols;
    chip_trans1->update_timestamp(current_time);
    chip_trans1->set_address(orig_trans->get_address());
    chip_trans1->set_command(orig_trans->get_command());
    chip_trans1->set_data_ptr(reinterpret_cast<unsigned char*>(&chip_trans1->data));
    chip_trans1->set_data_length(orig_trans->get_data_length());
    chip_trans1->set_streaming_width(orig_trans->get_streaming_width());
    chip_trans1->set_byte_enable_ptr(orig_trans->get_byte_enable_ptr());
    chip_trans1->set_dmi_allowed(orig_trans->is_dmi_allowed());

    int address = chip_trans1->get_address(); // address is used to determine wavelength
    if (address >= 0x100 && address < 0x200) {
        // Single-chiplet communication
        int chiplet_index = address - 0x100;
        int chiplet_id = chiplet_index - PE_PER_CHIPLET;
        
        int single_wavelength = chiplet_index;
        
        trans.set_response_status(TLM_OK_RESPONSE);  // Set response status
        //log_info(std::to_string(chiplet_id) + " Single-chiplet communication, wavelength: " + std::to_string(single_wavelength));
        
        // Single-chiplet communication (λM)
        log_info(std::to_string(chiplet_id) + " Received input feature: " + std::to_string(chip_trans1->data) + " with Wavelength: " + std::to_string(single_wavelength));
        //wait(sc_time(5, SC_NS));
        //sc_time delay = SC_ZERO_TIME;
        delay += sc_time(5, SC_NS);
        // Forward the transaction 
        forward_features_to_O_E_converters(*chip_trans1, delay, single_wavelength);

    } else {
        log_error(std::to_string(chiplet_id) + " Invalid or unrecognized address: " + std::to_string(address));
    }

    delete chip_trans1;  // Clean up dynamically allocated memory
    
}

void ChipletInterface::forward_features_to_O_E_converters(CustomPayload& trans, sc_time& delay, int single_wavelength){
    for (int i = 0; i < num_pes; ++i) {
        log_info(std::to_string(chiplet_id) + " Forwarding input features: " + std::to_string(trans.data) + " with Wavelength: " + std::to_string(single_wavelength) + " to Optical to Electrical Converter: " + std::to_string(i));
        oec_features_send_sockets[i] -> b_transport(trans, delay); // Send transaction to OEC    
    }
    if(trans.get_response_status() == TLM_OK_RESPONSE) {
        log_info( std::to_string(chiplet_id) + " Transaction of input feature " + std::to_string(trans.data) + " completed successfully from Chiplet Interface: " + std::to_string(chiplet_id) + " to OEC");
    } else {
        log_error(std::to_string(chiplet_id) +  " Error in transmitting data: " + std::to_string(trans.data) + " from Chiplet Interface to OEC");
    }
    
}


// Handle incoming transactions - weight kernals from the Interposer Interface
void ChipletInterface::ii_weight_receive_b_transport(tlm_generic_payload& trans, sc_time& delay) {
    CustomPayload* orig_trans = dynamic_cast<CustomPayload*>(&trans);
    if (!orig_trans) {
        
        log_error( std::to_string(chiplet_id) + " Invalid transaction type (not CustomPayload)." );
        trans.set_response_status(TLM_GENERIC_ERROR_RESPONSE);
        return;
    }

    CustomPayload* chip_trans2 = new CustomPayload();

    sc_time current_time = sc_time_stamp();
    //manually copy the transaction payload for future use
    // Log the time when the core sent the transaction
    sc_time sent_time = orig_trans->timestamp;
    chip_trans2->data = orig_trans->data;
    chip_trans2->num_rows = orig_trans->num_rows;
    chip_trans2->num_cols = orig_trans->num_cols;
    chip_trans2->update_timestamp(current_time);
    chip_trans2->set_address(orig_trans->get_address());
    chip_trans2->set_command(orig_trans->get_command());
    chip_trans2->set_data_ptr(reinterpret_cast<unsigned char*>(&chip_trans2->data));
    chip_trans2->set_data_length(orig_trans->get_data_length());
    chip_trans2->set_streaming_width(orig_trans->get_streaming_width());
    chip_trans2->set_byte_enable_ptr(orig_trans->get_byte_enable_ptr());
    chip_trans2->set_dmi_allowed(orig_trans->is_dmi_allowed());

    int address = chip_trans2->get_address(); // address is used to determine wavelength
    if (address >= 0x200) {        
        int cross_wavelength = address - 0x200;
        //log_info(std::to_string(chiplet_id) + " Cross-chiplet communication, wavelength: " + std::to_string(cross_wavelength));
        
        log_info(std::to_string(chiplet_id) + " Received Weight Kernal: " + std::to_string(chip_trans2->data) + " with Wavelength: " + std::to_string(cross_wavelength));
    
        // Forward the transaction to the chiplet interface
        //sc_time delay = SC_ZERO_TIME;
        delay += sc_time(5, SC_NS);
        forward_weight_to_O_E_converters(*chip_trans2, delay, cross_wavelength);
       
    } else {
        log_error(std::to_string(chiplet_id) + " Invalid or unrecognized address: " + std::to_string(address));
    }

    delete chip_trans2;  // Clean up dynamically allocated memory
    trans.set_response_status(TLM_OK_RESPONSE);  // Set response status
}

void ChipletInterface::forward_weight_to_O_E_converters(CustomPayload& trans, sc_time& delay, int cross_wavelength){
    oec_weight_send_sockets[cross_wavelength] -> b_transport(trans, delay); // Send transaction to OEC
    log_info(std::to_string(chiplet_id) + " Forwarding Weight Kernal: " + std::to_string(trans.data) + " with Wavelength: " + std::to_string(cross_wavelength) + " to Optical to Electrical Converter: " + std::to_string(cross_wavelength));
    if(trans.get_response_status() == TLM_OK_RESPONSE) {
        log_info( std::to_string(chiplet_id) + " Transaction of weight kernal" + std::to_string(trans.data) + " completed successfully from Chiplet Interface: " + std::to_string(chiplet_id) + " to OEC");
    } else {
        log_error(std::to_string(chiplet_id) +  " Error in transmitting data: " + std::to_string(trans.data) + " from Chiplet Interface to OEC");
    }
    
}

//Handle incoming transactions from the OEC
void ChipletInterface::oec_b_transport(tlm_generic_payload& trans, sc_time& delay) {
    PESendPayload* orig_trans = dynamic_cast<PESendPayload*>(&trans);
    if (!orig_trans) {
        
        log_error(std::to_string(chiplet_id) + "  Invalid transaction type (not PESendPayload).");
        trans.set_response_status(TLM_GENERIC_ERROR_RESPONSE);
        return;
    }

    PESendPayload* rx_trans2 = new PESendPayload();

    sc_time current_time = sc_time_stamp();
    //manually copy the transaction payload for future use
    // Log the time when the core sent the transaction
    sc_time sent_time = orig_trans->timestamp;
    rx_trans2->psum = orig_trans->psum;
    rx_trans2->src_chiplet = orig_trans->src_chiplet;
    rx_trans2->src_pe = orig_trans->src_pe;
    rx_trans2->num_rows = orig_trans->num_rows;
    rx_trans2->num_cols = orig_trans->num_cols;
    rx_trans2->update_timestamp(current_time);
    rx_trans2->set_address(orig_trans->get_address());
    rx_trans2->set_command(orig_trans->get_command());
    rx_trans2->set_data_ptr(reinterpret_cast<unsigned char*>(&rx_trans2->psum));
    rx_trans2->set_data_length(orig_trans->get_data_length());
    rx_trans2->set_streaming_width(orig_trans->get_streaming_width());
    rx_trans2->set_byte_enable_ptr(orig_trans->get_byte_enable_ptr());
    rx_trans2->set_dmi_allowed(orig_trans->is_dmi_allowed());

    int address = rx_trans2->get_address(); // address is used to determine wavelength

    int chiplet_index = address - 0x300;
   
    int unicast_wavelength = chiplet_index;
    
    log_info(std::to_string(chiplet_id) + " Chiplet Interface received PSUM from PE: " + std::to_string(rx_trans2->src_pe) + " through OEC with data: " + std::to_string(rx_trans2->psum));
    trans.set_response_status(TLM_OK_RESPONSE);  // Set response status
    // Forward the transaction to the Interposer interface 
    //delay = SC_ZERO_TIME;
    delay += sc_time(5, SC_NS);
    forward_psum_to_interposer(*rx_trans2, delay);
       
    
    delete rx_trans2;  // Clean up dynamically allocated memory
    
}

void ChipletInterface::forward_psum_to_interposer(PESendPayload& trans, sc_time& delay){
    log_info(std::to_string(chiplet_id) + " Forwarding the PSUM " + std::to_string(trans.psum) + " from PE: " + std::to_string(trans.src_pe) + " to Interposer interface: " + std::to_string(trans.src_chiplet));
    interposer_send_socket->b_transport(trans, delay); 
    if(trans.get_response_status() == TLM_OK_RESPONSE) {
        log_info(std::to_string(chiplet_id) + " PSUM Transaction: " + std::to_string(trans.psum) + " completed successfully from Chiplet Interface: " + std::to_string(chiplet_id) + " to Interposer Interface");
    }else{
        log_error( std::to_string(chiplet_id) + " Error in transmitting PSUM: " + std::to_string(trans.psum) + " from Chiplet Interface to Interposer Interface ");
    }
    
}

void ChipletInterface::log_info(std::string msg){
    SC_REPORT_INFO(C_LOG, msg.c_str());
}

void ChipletInterface::log_error(std::string msg){
    SC_REPORT_ERROR(C_LOG, msg.c_str());
}