#include "Global_Buffer.h"

// Constructor implementation
GlobalBuffer::GlobalBuffer(sc_module_name name, int num_chiplets, int num_pes)
    : sc_module(name),
      num_chiplets(num_chiplets),
      num_pes(num_pes), 
      feature_counter(num_chiplets, 20), 
      weight_counter(num_pes, 20), 
      credit_matrix(num_chiplets, std::vector<int>(num_pes, 0)) {
    //sc_report_handler::set_log_file_name("out/report.log");
    
    sc_report_handler::set_actions(GB_LOG, SC_INFO, SC_LOG|SC_DISPLAY);
    for (int i = 0; i < num_chiplets; ++i) {
        // Register b_transport callback for each target socket
        receive_to_GB_sockets[i].register_b_transport(this, &GlobalBuffer::receive_to_gb_b_transport);
    }
  

    // SC_THREAD for processing features, weights, and PSum reception
    SC_THREAD(single_chiplet_communication);
    SC_THREAD(cross_chiplet_communication);
    
}

// Generate input features and transmit as single-chiplet communication
void GlobalBuffer::generate_features() {
    CustomPayload* trans = nullptr;
    // Initialize random seed 
    static bool seeded1 = false;
    if (!seeded1) {
        srand(time(0));  // Seed the random number generator
        seeded1 = true;
    }
    sc_time inter_transaction_delay = sc_time(10, SC_NS);  // Delay between transactions
    sc_core::sc_time delay = SC_ZERO_TIME;
    bool retry = false;  // Retry flag
    
    // Iterate over all chiplets
    for (int chiplet_id = 0; chiplet_id < CHIPLET_NO; ++chiplet_id) {
        do {  // Ensure feature counter > 0
            if (!retry) {
                int data = rand() % 101;
                unsigned int num_rows = rand() % 101;
                unsigned int num_cols = rand() % 101;
                log_info("Generated input feature: " + std::to_string(data));
                // Create a new payload for each chiplet
                trans = new CustomPayload(num_rows, num_cols, data, sc_time_stamp());
                int chiplet_index = 0x100 + (PE_PER_CHIPLET + chiplet_id);  // Address for wavelength λ8 onward
                trans->update_timestamp(sc_time_stamp());  // Set the current simulation time
                // Set TLM payload fields
                trans->set_command(TLM_WRITE_COMMAND);
                trans->set_address(chiplet_index);
                trans->set_byte_enable_ptr(nullptr);
                trans->set_dmi_allowed(false);
            }           
            //std::cout << "Attempt transmitting input features" << std::endl;
            // Attempt transaction
            if (feature_counter[chiplet_id] > 0) {
                //std::cout << "Transmitting input features" << std::endl;
                log_info("Transmitting Input Features: " + std::to_string(trans->data));
                //wait(sc_time(5, SC_NS));
                sc_time start_time = sc_time_stamp();
                //std::cout << "delay before send in Global buffer" << start_time << std::endl;
                features_from_GB_sockets[chiplet_id]->b_transport(*trans, delay);
                sc_time end_time = sc_time_stamp();
                //std::cout << "delay before send in Global buffer" << end_time << std::endl;
                if(trans->get_response_status() == TLM_OK_RESPONSE) {
                    log_info("Transaction of input feature " + std::to_string(trans->data) + " completed successfully from Global Buffer to Interpoaser Interface");
                } else {
                    log_error ("Error in transmitting data: " + std::to_string(trans->data) + " from Global Buffer to Interposer Interface");
                }
                retry = false;  // Transaction successful, no retry needed
                feature_counter[chiplet_id]--;  // Decrement feature counter
                delete trans;  // Clean up payload
                wait(inter_transaction_delay);
                break;  // Exit retry loop
            } else {
                retry = true;  // FIFO full, retry needed
                log_info("GlobalBuffer: FIFO full, retrying...");
                wait(20, SC_NS);  // Wait for FIFO space
            }

            if (trans != nullptr) {
                delete trans;  // Clean up if not already deleted
            }
        } while(retry);
    }

    wait(SC_ZERO_TIME);  // Synchronize simulation time
}


// Process to transmit weight kernels
void GlobalBuffer::generate_weights() {
    CustomPayload* trans = nullptr;
    static bool seeded = false;
    if (!seeded) {
        srand(time(0));  // Seed the random number generator
        seeded = true;
    }
    sc_time delay = SC_ZERO_TIME;  // Start with zero delay
    sc_time inter_transaction_delay = sc_time(15, SC_NS);  // Delay between transactions
    bool retry = false;  // Retry flag
    for (int pe_id = 0; pe_id < PE_PER_CHIPLET; ++pe_id) {
        do {
            if (!retry) {
                int data = rand() % 101;
                unsigned int num_rows = rand() % 101;
                unsigned int num_cols = rand() % 101;
                log_info("Generated Weight kernal: " + std::to_string(data));
                // Create a new payload for each chiplet
                trans = new CustomPayload(num_rows, num_cols, data, sc_time_stamp());
                int chiplet_index = 0x200 + (pe_id);  // Address for wavelength λ8 onward
                // Set TLM payload fields
                trans->set_command(TLM_WRITE_COMMAND);
                trans->set_address(chiplet_index);
                
                trans->set_byte_enable_ptr(nullptr);
                trans->set_dmi_allowed(false);
            }
            //std::cout << "attempt transmitting weights" << std::endl;
            for (int chiplet_id = 0; chiplet_id < num_chiplets; ++chiplet_id) {
                //std::cout << "Transmitting weights" << std::endl;        
                if (weight_counter[pe_id] > 0){
                    log_info("Transmitting weight kernals: " + std::to_string(trans->data) + " to interposer Interface " + std::to_string(chiplet_id));
                    (weight_from_GB_sockets[chiplet_id])->b_transport(*trans, delay);
                    if(trans->get_response_status() == TLM_OK_RESPONSE) {
                        log_info("Transaction of Weight Kernal " + std::to_string(trans->data) + " completed successfully from Global Buffer: " +" to Interpoaser Interface");
                    }else{
                        log_error ("Error in transmitting weight kernal: " + std::to_string(trans->data) + " from Global Buffer to Interposer Interface");
                    }
                    weight_counter[pe_id]--;            	
                    retry = false;
                    
                } else {
                    retry = true;  // FIFO full, retry needed
                    log_info("Waiting for space in FIFO for weight kernals...");
                    wait(20, SC_NS);  // Wait before retrying
                            
                }
            } wait(inter_transaction_delay);

        } while(retry);
        
    }
    wait(SC_ZERO_TIME);  // Synchronize simulation time
}

// Process to receive PSums from all chiplets
void GlobalBuffer::receive_to_gb_b_transport(tlm_generic_payload& trans, sc_time& delay) {
    PESendPayload* orig_trans = dynamic_cast<PESendPayload*>(&trans);
    if (!orig_trans) {
        log_error("  Invalid transaction type (not PESendPayload).");
        trans.set_response_status(TLM_GENERIC_ERROR_RESPONSE);
        return;
    }

    PESendPayload* rx_trans4 = new PESendPayload();

    sc_time current_time = sc_time_stamp();
    //manually copy the transaction payload for future use
    // Log the time when the core sent the transaction
    sc_time sent_time = orig_trans->timestamp;
    rx_trans4->psum = orig_trans->psum;
    rx_trans4->src_chiplet = orig_trans->src_chiplet;
    unsigned int chiplet_id = rx_trans4->src_chiplet;
    rx_trans4->src_pe = orig_trans->src_pe;
    unsigned int pe_id = rx_trans4->src_pe;
    rx_trans4->num_rows = orig_trans->num_rows;
    rx_trans4->num_cols = orig_trans->num_cols;
    rx_trans4->update_timestamp(current_time);
    rx_trans4->set_address(orig_trans->get_address());
    rx_trans4->set_command(orig_trans->get_command());
    rx_trans4->set_data_ptr(reinterpret_cast<unsigned char*>(&rx_trans4->psum));
    rx_trans4->set_data_length(orig_trans->get_data_length());
    rx_trans4->set_streaming_width(orig_trans->get_streaming_width());
    rx_trans4->set_byte_enable_ptr(orig_trans->get_byte_enable_ptr());
    rx_trans4->set_dmi_allowed(orig_trans->is_dmi_allowed());

    int address = rx_trans4->get_address(); // address is used to determine wavelength

    int chiplet_index = address - 0x300;
    int unicast_wavelength = chiplet_index;
    // wait(4, SC_NS);
    log_info(" Received transaction from PE: " + std::to_string(rx_trans4->src_pe) + " from chiplet: " + std::to_string(rx_trans4->src_chiplet) + " through Interposer Interface with PSUM: " + std::to_string(rx_trans4->psum));
    // Increment the credit matrix value
    credit_matrix[chiplet_id][pe_id]++;
    log_info("Updated credit_matrix[" + std::to_string(chiplet_id) + "][" + std::to_string(pe_id) + "] = " +
                  std::to_string(credit_matrix[chiplet_id][pe_id]) );
    //log_credits(rx_trans4);
    // Check and update weight counter
    bool all_pe_in_column_nonzero = true;
    for (int i = 0; i < num_chiplets; ++i) {
        if (credit_matrix[i][pe_id] == 0) {
            all_pe_in_column_nonzero = false;
            break;
        }
    }
    if (all_pe_in_column_nonzero) {
        weight_counter[pe_id]++;
        log_info( "Incremented weight_counter[" + std::to_string( pe_id ) +  "] = " +std::to_string( weight_counter[pe_id]));
    }


    // Check and update feature counter
    bool all_pe_in_row_nonzero = true;
    for (int j = 0; j < num_pes; ++j) {
        if (credit_matrix[chiplet_id][j] == 0) {
            all_pe_in_row_nonzero = false;
            break;
        }
    }
    if (all_pe_in_row_nonzero) {
        feature_counter[chiplet_id]++;
        log_info("Incremented feature_counter[" + std::to_string(chiplet_id) + "] = " + std::to_string(feature_counter[chiplet_id]));
    }

    delete rx_trans4;  // Clean up dynamically allocated memory
    trans.set_response_status(TLM_OK_RESPONSE);  // Set response status
}

// Single-chiplet communication (λM onwards for input features)
void GlobalBuffer::single_chiplet_communication() {
    while(true){
        
        generate_features();
        
        wait();
    }
}

// Cross-chiplet communication (λ0 to λ(M-1) for weights)
void GlobalBuffer::cross_chiplet_communication() {
    while(true){
        
        generate_weights();
        
        wait();
    }    
}



void GlobalBuffer::log_info(std::string msg){
    SC_REPORT_INFO(GB_LOG, msg.c_str());
}

void GlobalBuffer::log_error(std::string msg){
    SC_REPORT_ERROR(GB_LOG, msg.c_str());
}