#include "Interposer_Interface.h"
#include <iostream>

// Constructor implementation
InterposerInterface::InterposerInterface(sc_module_name name)
    : sc_module(name)
      //chiplet_receive_socket("chiplet_receive_socket"),
      //chiplet_send_socket("chiplet_send_socket") 
    {
    //sc_report_handler::set_log_file_name("out/report.log");
    sc_report_handler::set_actions(I_LOG, SC_INFO, SC_LOG|SC_DISPLAY);
    // Register b_transport methods for GB and Chiplet interfaces
    gb_features_receive_socket.register_b_transport(this, &InterposerInterface::gb_features_b_transport);
    gb_weight_receive_socket.register_b_transport(this, &InterposerInterface::gb_weight_b_transport);
    chiplet_receive_socket.register_b_transport(this, &InterposerInterface::chiplet_b_transport);
    
}

// Handle incoming transactions from the Global Buffer
void InterposerInterface::gb_features_b_transport(tlm_generic_payload& trans, sc_time& delay) {
    CustomPayload* orig_trans = dynamic_cast<CustomPayload*>(&trans);
    if (!orig_trans) {
        wait(5, SC_NS);
        log_error("Invalid transaction type (not CustomPayload).");
        trans.set_response_status(TLM_GENERIC_ERROR_RESPONSE);
        return;
    }


    // Copy payload manually
    CustomPayload* my_trans = new CustomPayload();
    sc_time current_time = sc_time_stamp();
    my_trans->update_timestamp(sc_time_stamp());
    // Log the time when the core sent the transaction
    sc_time sent_time = orig_trans->timestamp;
    my_trans->data = orig_trans->data;
    my_trans->num_rows = orig_trans->num_rows;
    my_trans->num_cols = orig_trans->num_cols;
    my_trans->update_timestamp(current_time);
    my_trans->set_address(orig_trans->get_address());
    my_trans->set_command(orig_trans->get_command());
    my_trans->set_data_ptr(reinterpret_cast<unsigned char*>(&my_trans->data));
    my_trans->set_data_length(orig_trans->get_data_length());
    my_trans->set_streaming_width(orig_trans->get_streaming_width());
    my_trans->set_byte_enable_ptr(orig_trans->get_byte_enable_ptr());
    my_trans->set_dmi_allowed(orig_trans->is_dmi_allowed());
    
    int address = my_trans->get_address(); // address is used to determine wavelength
    if (address >= 0x100 && address < 0x200) {
        // Single-chiplet communication
        int chiplet_index = address - 0x100;
        int chiplet_id = chiplet_index - PE_PER_CHIPLET;  
        int single_wavelength = chiplet_index;
        trans.set_response_status(TLM_OK_RESPONSE);  // Set response status

        log_info(std::to_string(interposer_id) + " Received input feature: " + std::to_string(my_trans->data) + " with Wavelength: " + std::to_string(single_wavelength));
        //wait(sc_time(5, SC_NS));
        //sc_time delay = SC_ZERO_TIME;
        delay += sc_time(5, SC_NS);
        // Forward the transaction to the chiplet interface
        chiplet_features_send_transaction(*my_trans, delay, single_wavelength);
        
    } else {
        log_error(std::to_string(interposer_id) + " Invalid or unrecognized address: " + std::to_string(address));
    }

    delete my_trans;  // Clean up dynamically allocated memory
    
}

void InterposerInterface::chiplet_features_send_transaction(CustomPayload& trans, sc_time& delay, int single_wavelength) {
    // Forward the transaction to the chiplet interface
    log_info(std::to_string(interposer_id) + " Forwarding input features: " + std::to_string(trans.data) + " with Wavelength: " + std::to_string(single_wavelength) + " to chiplet interface");
    chiplet_features_send_socket->b_transport(trans, delay);  // Send transaction to Chiplet Interface
    
    if(trans.get_response_status() == tlm::TLM_OK_RESPONSE) {
        log_info(std::to_string(interposer_id) + " Transaction of input feature " + std::to_string(trans.data) + " completed successfully from Interposer Interface: " + std::to_string(interposer_id) + " to Chiplet Interface");
    } else {
        log_error(std::to_string(interposer_id) + " Error in transmitting data: " + std::to_string(trans.data) + " from Interposer Interface to Chiplet Interface");
    }
    
}

// Handle incoming weight kernals from the Global Buffer
void InterposerInterface::gb_weight_b_transport(tlm_generic_payload& trans, sc_time& delay) {
    CustomPayload* orig_trans = dynamic_cast<CustomPayload*>(&trans);
    if (!orig_trans) {
        // wait(4, SC_NS);
        std::cerr << std::to_string(interposer_id) + " Invalid transaction type (not CustomPayload)." << std::endl;
        trans.set_response_status(TLM_GENERIC_ERROR_RESPONSE);
        return;
    }

  CustomPayload* new_trans = new CustomPayload();
  sc_time current_time = sc_time_stamp();
    //manually copy the transaction payload for future use
    // Log the time when the core sent the transaction
    sc_time sent_time = orig_trans->timestamp;
    new_trans->data = orig_trans->data;
    new_trans->num_rows = orig_trans->num_rows;
    new_trans->num_cols = orig_trans->num_cols;
    new_trans->update_timestamp(current_time);
    new_trans->set_address(orig_trans->get_address());
    new_trans->set_command(orig_trans->get_command());
    new_trans->set_data_ptr(reinterpret_cast<unsigned char*>(&new_trans->data));
    new_trans->set_data_length(orig_trans->get_data_length());
    new_trans->set_streaming_width(orig_trans->get_streaming_width());
    new_trans->set_byte_enable_ptr(orig_trans->get_byte_enable_ptr());
    new_trans->set_dmi_allowed(orig_trans->is_dmi_allowed());

    int address = new_trans->get_address(); // address is used to determine wavelength
    if (address >= 0x200) {
        int cross_wavelength = address - 0x200;
        //log_info("Cross-chiplet communication, wavelength : " + std::to_string(cross_wavelength));
        // wait(4, SC_NS);
        // Cross-chiplet communication (λ0, λ1, λ2, ...)
        log_info(std::to_string(interposer_id) + " Received cross-chiplet data: " + std::to_string(new_trans->data) + " with Wavelength: " + std::to_string(cross_wavelength));
        trans.set_response_status(TLM_OK_RESPONSE);        
        log_info(std::to_string(interposer_id) + " Splitting data for PE: " + std::to_string(cross_wavelength) + " to chiplet:" + std::to_string(interposer_id));
        
        // Forward modified data to chiplet interface
        //sc_time delay = SC_ZERO_TIME;
        delay += sc_time(5, SC_NS);
        chiplet_weight_send_transaction(*new_trans, delay, cross_wavelength);
    } else {
        log_error(std::to_string(interposer_id) + " Invalid or unrecognized address: " + std::to_string(address));    
    }

    delete new_trans;  // Clean up dynamically allocated memory
    // Complete the original transaction    
}

void InterposerInterface::chiplet_weight_send_transaction(CustomPayload& trans, sc_time& delay, int cross_wavelength){

    chiplet_weight_send_socket->b_transport(trans, delay);  // Send transaction to Chiplet Interface
    log_info(std::to_string(interposer_id) + " Forwarding weight kernals: " + std::to_string(trans.data) + " with Wavelength: " + std::to_string(cross_wavelength) + " to chiplet interface");
    if(trans.get_response_status() == TLM_OK_RESPONSE) {
        log_info(std::to_string(interposer_id) + " Transaction of weight kernal " + std::to_string(trans.data) + " completed successfully from Interposer Interface: " + std::to_string(interposer_id) + " to Chiplet Interface");
    } else {
        log_error(std::to_string(interposer_id) + " Error in transmitting data: " + std::to_string(trans.data) + " from Interposer Interface to Chiplet Interface");
    }
    
}

//Handle incoming transactions from the Chiplet Interface
void InterposerInterface::chiplet_b_transport(tlm_generic_payload& trans, sc_time& delay) {
    PESendPayload* orig_trans = dynamic_cast<PESendPayload*>(&trans);
    if (!orig_trans) {
        log_error(std::to_string(interposer_id) + "  Invalid transaction type (not PESendPayload).");
        trans.set_response_status(TLM_GENERIC_ERROR_RESPONSE);
        return;
    }

    PESendPayload* rx_trans3 = new PESendPayload();

    sc_time current_time = sc_time_stamp();
    //manually copy the transaction payload for future use
    // Log the time when the core sent the transaction
    sc_time sent_time = orig_trans->timestamp;
    rx_trans3->psum = orig_trans->psum;
    rx_trans3->src_chiplet = orig_trans->src_chiplet;
    rx_trans3->src_pe = orig_trans->src_pe;
    rx_trans3->num_rows = orig_trans->num_rows;
    rx_trans3->num_cols = orig_trans->num_cols;
    rx_trans3->update_timestamp(current_time);
    rx_trans3->set_address(orig_trans->get_address());
    rx_trans3->set_command(orig_trans->get_command());
    rx_trans3->set_data_ptr(reinterpret_cast<unsigned char*>(&rx_trans3->psum));
    rx_trans3->set_data_length(orig_trans->get_data_length());
    rx_trans3->set_streaming_width(orig_trans->get_streaming_width());
    rx_trans3->set_byte_enable_ptr(orig_trans->get_byte_enable_ptr());
    rx_trans3->set_dmi_allowed(orig_trans->is_dmi_allowed());

    int address = rx_trans3->get_address(); // address is used to determine wavelength

    int chiplet_index = address - 0x300;
   
    int unicast_wavelength = chiplet_index;

    // wait(4, SC_NS);

    log_info(std::to_string(interposer_id) + " received PSUM from PE: " + std::to_string(rx_trans3->src_pe) + " through OEC with data: " + std::to_string(rx_trans3->psum));
    trans.set_response_status(TLM_OK_RESPONSE);  // Set response status

    // Forward the transaction to the Global Buffer 
    //delay = SC_ZERO_TIME;
    delay += sc_time(5, SC_NS);
    gb_send_transaction(*rx_trans3, delay);

    delete rx_trans3;  // Clean up dynamically allocated memory   
}

void InterposerInterface::gb_send_transaction(PESendPayload& trans, sc_time& delay){
    gb_send_socket->b_transport(trans, delay);    
    log_info(std::to_string(interposer_id) + " Forwarding the PSUMs from PE: " + std::to_string(trans.src_pe) + " to Global Buffer ");
    if(trans.get_response_status() == TLM_OK_RESPONSE) {
        log_info( std::to_string(interposer_id) + " Transaction of PSUM " + std::to_string(trans.psum) + "  completed successfully from Interposer Interface: " + std::to_string(interposer_id) + " to Global Buffer");
    }else{
        log_error( std::to_string(interposer_id) + "Error in transmitting PSUM: " + std::to_string(trans.psum) + " from Interposer Interface to Global Buffer");
    }
    
}

void InterposerInterface::log_info(std::string msg){
    SC_REPORT_INFO(I_LOG, msg.c_str());
}

void InterposerInterface::log_error(std::string msg){
    SC_REPORT_ERROR(I_LOG, msg.c_str());
}