#include "Processing_Element.h"

// PE module with FIFOs, semaphores, and capacity control for receive FIFO
ProcessingElement::ProcessingElement(sc_module_name name, TokenManager& token_ref)
: sc_module(name), token(token_ref), features_fifo(IFMAP_FIFO_CAPACITY), weight_fifo(WEIGHT_FIFO_CAPACITY), psum_fifo(ACC_FIFO_CAPACITY), pe_id(-1) // default ProcessingElement id
{
    //sc_report_handler::set_log_file_name("out/report.log");
    sc_report_handler::set_actions(PE_LOG, SC_INFO, SC_LOG|SC_DISPLAY);
    
    oec_features_receive_socket.register_b_transport(this, &ProcessingElement::oec_features_b_transport); //register the b_transport fun to each target socket to receive data from oec
    oec_weight_receive_socket.register_b_transport(this, &ProcessingElement::oec_weight_b_transport);
       
    SC_THREAD(process_convolution); //process to handle data in the input fifo 
    SC_THREAD(process_psum_send); //process to send the data available in output fifo to destination 
}



void ProcessingElement::oec_features_b_transport(tlm_generic_payload& trans, sc_time& delay) {
    CustomPayload* orig_trans = dynamic_cast<CustomPayload*>(&trans);
    if (!orig_trans) {
        // wait(4, SC_NS);
        log_error(std::to_string(pe_id) + " Invalid transaction type (not CustomPayload).");
        trans.set_response_status(TLM_GENERIC_ERROR_RESPONSE);
        return;
    }

    CustomPayload* pe_trans1 = new CustomPayload();

    sc_time current_time = sc_time_stamp();
    //manually copy the transaction payload for future use
    // Log the time when the core sent the transaction
    sc_time sent_time = orig_trans->timestamp;
    pe_trans1->data = orig_trans->data;
    pe_trans1->num_rows = orig_trans->num_rows;
    pe_trans1->num_cols = orig_trans->num_cols;
    pe_trans1->update_timestamp(current_time);
    pe_trans1->set_address(orig_trans->get_address());
    pe_trans1->set_command(orig_trans->get_command());
    pe_trans1->set_data_ptr(reinterpret_cast<unsigned char*>(&pe_trans1->data));
    pe_trans1->set_data_length(orig_trans->get_data_length());
    pe_trans1->set_streaming_width(orig_trans->get_streaming_width());
    pe_trans1->set_byte_enable_ptr(orig_trans->get_byte_enable_ptr());
    pe_trans1->set_dmi_allowed(orig_trans->is_dmi_allowed());

    int address = pe_trans1->get_address(); // address is used to determine wavelength
    if (address >= 0x100 && address < 0x200) {
        // Single-chiplet communication
        int chiplet_index = address - 0x100;
        int chiplet_id = chiplet_index - PE_PER_CHIPLET;
        
        int single_wavelength = chiplet_index;
        //log_info(std::to_string(pe_id) + " Single-chiplet communication, wavelength: " + std::to_string(single_wavelength));
        // wait(4, SC_NS);
        // Single-chiplet communication (λM)
        log_info(std::to_string(pe_id) + " Received input feature: " + std::to_string(pe_trans1->data) + " with Wavelength: " + std::to_string(single_wavelength));
        delay += sc_time(5, SC_NS);
        features_fifo.write(pe_trans1 -> data); //Write input features to fifo
        
        log_info(std::to_string(pe_id) + " writing the  input features: " + std::to_string(pe_trans1->data) + " with Wavelength: " + std::to_string(single_wavelength) + " to FiFo");
    
    } else {

        log_error(std::to_string(pe_id) + " Invalid or unrecognized address: " + std::to_string(address));
    
    }

    delete pe_trans1;  // Clean up dynamically allocated memory
    trans.set_response_status(TLM_OK_RESPONSE);  // Set response status
}


// Handle incoming weight kernals from the Global Buffer
void ProcessingElement::oec_weight_b_transport(tlm_generic_payload& trans, sc_time& delay) {
    CustomPayload* orig_trans = dynamic_cast<CustomPayload*>(&trans);
    if (!orig_trans) {
        // wait(4, SC_NS);
        log_error(std::to_string(pe_id) + " Invalid transaction type (not CustomPayload).");
        trans.set_response_status(TLM_GENERIC_ERROR_RESPONSE);
        return;
    }

  CustomPayload* pe_trans2 = new CustomPayload();
  sc_time current_time = sc_time_stamp();
    //manually copy the transaction payload for future use
    // Log the time when the core sent the transaction
    sc_time sent_time = orig_trans->timestamp;
    pe_trans2->data = orig_trans->data;
    pe_trans2->num_rows = orig_trans->num_rows;
    pe_trans2->num_cols = orig_trans->num_cols;
    pe_trans2->update_timestamp(current_time);
    pe_trans2->set_address(orig_trans->get_address());
    pe_trans2->set_command(orig_trans->get_command());
    pe_trans2->set_data_ptr(reinterpret_cast<unsigned char*>(&pe_trans2->data));
    pe_trans2->set_data_length(orig_trans->get_data_length());
    pe_trans2->set_streaming_width(orig_trans->get_streaming_width());
    pe_trans2->set_byte_enable_ptr(orig_trans->get_byte_enable_ptr());
    pe_trans2->set_dmi_allowed(orig_trans->is_dmi_allowed());

    int address = pe_trans2->get_address(); // address is used to determine wavelength
    if (address >= 0x200) {
    int cross_wavelength = address - 0x200;
    //log_info("Cross-chiplet communication, wavelength : " + std::to_string(cross_wavelength));
    // wait(4, SC_NS);
    // Cross-chiplet communication (λ0, λ1, λ2, ...)
    log_info(std::to_string(pe_id) + " Received cross-chiplet data: " + std::to_string(pe_trans2->data) + " with Wavelength: " + std::to_string(cross_wavelength));
    delay += sc_time(5, SC_NS);
    weight_fifo.write(pe_trans2-> data); //Write weights to fifo

    log_info(std::to_string(pe_id) + " Saving weight kernals: " + std::to_string(pe_trans2->data) + " with Wavelength: " + std::to_string(cross_wavelength) + " to Fifo");
   
    } else {

        log_error(std::to_string(pe_id) + " Invalid or unrecognized address: " + std::to_string(address));
        
    }

    delete pe_trans2;  // Clean up dynamically allocated memory
    // Complete the original transaction
    trans.set_response_status(TLM_OK_RESPONSE);
}




// Thread to process input fifo
void ProcessingElement::process_convolution() {
    
    while (true) {

        // Wait for data in the input FIFO
        int features = features_fifo.read();  // Blocking read
        int weight = weight_fifo.read();  // Blocking read
        int psum = features * weight;
        psum_fifo.write(psum);

    } 
}    
    

// Thread to process ProcessingElement_psum_fifo and forward transactions to gb
void ProcessingElement::process_psum_send() {    
    PESendPayload* trans = nullptr;
    sc_core::sc_time prev_timestamp = sc_time(0.1, SC_NS);
    while (true) {
        bool success = false;
        while (!success) {
            int psum = psum_fifo.read();  // Blocking read 
            if (token.request_token(pe_id, pe_chiplet_id))  {
                unsigned int src_chiplet = pe_chiplet_id;
                unsigned int src_pe = pe_id;
                unsigned int num_rows = rand() % 5;
                unsigned int num_cols = rand() % 5;
                // Create a new payload for each chiplet
                trans = new PESendPayload(src_chiplet, src_pe, num_rows, num_cols, psum, sc_time_stamp());
                int chiplet_index = 0x300 + (pe_chiplet_id);  // Address for wavelength λ8 onward
                trans->update_timestamp(sc_time_stamp());  // Set the current simulation time
                // Set TLM payload fields
                trans->set_command(TLM_WRITE_COMMAND);
                trans->set_address(chiplet_index);
                trans->set_byte_enable_ptr(nullptr);
                trans->set_dmi_allowed(false);
                sc_time delay = SC_ZERO_TIME;
                // wait(4, SC_NS);
                // Attempt transaction               
                log_info(std::to_string(pe_id) + " Transmitting PSUM: " + std::to_string(psum));
                oec_send_socket ->b_transport(*trans, delay);
                if(trans->get_response_status() == TLM_OK_RESPONSE) {
                    log_info( std::to_string(pe_id) + " Transaction of PSUM completed successfully at time " + sc_time_stamp().to_string() + " from PE: " + std::to_string(pe_id) + " to OEC");
                }else{
                    log_error(std::to_string(pe_id) +  " Error in transmitting PSUM: " + std::to_string(psum) + " from PE to OEC ");
                }
                token.release_token(pe_id, pe_chiplet_id);
                success = true;
            } else{
                log_info(std::to_string(pe_id) + ": Token of the Chiplet" + std::to_string(pe_chiplet_id) + " is unavailable. Retrying with data: " + std::to_string(psum));
                wait(sc_time(5, SC_NS));  // Wait before retrying
            }
        }        
    }
}
    
//print log messages
void ProcessingElement::log_info(std::string msg){
    SC_REPORT_INFO(PE_LOG, msg.c_str());
}

void ProcessingElement::log_error(std::string msg){
    SC_REPORT_ERROR(PE_LOG, msg.c_str());
}