/*******************************************************************************
 * Copyright (C) 2024
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 ******************************************************************************/
// token_manager.h
#ifndef TOKEN_MANAGER_H
#define TOKEN_MANAGER_H

#include <systemc>
#include <vector>
#include <queue>
#include <iostream>
#include <mutex>
#include <unordered_set>
#include "configuration.h"

using namespace sc_core;

class TokenManager : public sc_module {
public:
    SC_HAS_PROCESS(TokenManager);

    TokenManager(sc_module_name name) : sc_module(name) {
        token_owner =- 1;      // Start with all tokens unowned
        request_queues.resize(PE_PER_CHIPLET);         // Initialize a queue for each token
        queue_entries.resize(PE_PER_CHIPLET);           // Set for each token to track queued PEs
        token_status = false; // All tokens are initially free
        last_pe_owner = -1; // Initialize last owner to -1
        sc_report_handler::set_log_file_name("out/report.log");
        sc_report_handler::set_actions(LOG_NAME, SC_INFO, SC_LOG|SC_DISPLAY);
        // Declare the token for access control
        //sc_core::sc_semaphore sem;
    }

    // Request a token for a specific PE with round-robin priority ordering
    bool request_token(int pe_id, int token_id) {
        //wait(SC_ZERO_TIME);
        std::lock_guard<sc_mutex> lock(mutex);  // Ensure atomic access
        if (token_id >= 0 && token_id < CHIPLET_NO) {
            wait(SC_ZERO_TIME);
            // Give priority as the request follows round-robin order from the last owner
            if (token_owner == pe_id || (token_owner == -1 && request_queues[pe_id].empty()))
                {
                log_info("Token for the chiplet " + std::to_string(token_id) + " granted to PE " + std::to_string(pe_id));
                token_owner = pe_id;
                token_status = true;  // Mark token as in use
                last_pe_owner = pe_id; // Update last owner
                return true;
            } else {
               if (queue_entries[pe_id].find(pe_id) == queue_entries[pe_id].end()) {
                    request_queues[pe_id].push(pe_id);
                    queue_entries[pe_id].insert(pe_id);  // Track in the set
                    log_info("PE " + std::to_string(pe_id) + " added to queue for token " + std::to_string(token_id));
                }
            }
        }
        return false;
    }

    // Release the token and pass it to the next in round-robin order
    bool release_token(int pe_id, int token_id) {
        std::lock_guard<sc_mutex> lock(mutex);  // Ensure atomic access
        if (token_id >= 0 && token_id < CHIPLET_NO) {
            if (token_owner == pe_id) {
                log_info("PE " + std::to_string(pe_id) + " released token for the chiplet " + std::to_string(token_id));

                if (!request_queues[token_id].empty()) {
                // Start searching from the next pe in clockwise order
                int next_pe = -1;
                int start_pe = (pe_id + 1) % CHIPLET_NO;  // Start clockwise from the next pe

                // Search for the first pe in queue that is in clockwise order
                for (int i = 0; i < CHIPLET_NO; ++i) {
                    int candidate_pe = (start_pe + i) % CHIPLET_NO;

                    // Check if this candidate is in the queue
                    if (queue_entries[pe_id].find(candidate_pe) != queue_entries[pe_id].end()) {
                        next_pe = candidate_pe;
                        break;
                    }
                }

                if (next_pe != -1) {
                    // Assign the token to the found pe
                    token_owner = next_pe;
                    token_status = true;  // Mark token as in use

                    // Remove next_pe from the queue and queue_entries
                    std::queue<int> temp_queue;
                    while (!request_queues[pe_id].empty()) {
                        int current_pe = request_queues[pe_id].front();
                        request_queues[pe_id].pop();

                        if (current_pe != next_pe) {
                            temp_queue.push(current_pe);  // Keep all others in the queue
                        }
                    }
                    request_queues[pe_id] = temp_queue;
                    queue_entries[pe_id].erase(next_pe);  // Remove from set

                    log_info("Token of the Chiplet " + std::to_string(token_id) + " passed to PE " + std::to_string(next_pe));
                }
            } else {
                token_owner= -1;  // No one else is waiting, so free the token
                token_status = false; // Mark token as free
            }
            return true;
        }
    }
    return false;
}

bool peek_token(int pe_id, int token_id) {
    if(token_status == false)
    {
        return true;
    }
    return false;
}
 
private:
    int token_owner;                // Tracks pe ownership (-1 means no owner)
    std::vector<std::queue<int>> request_queues;      // Queue of pending requests per token
    bool token_status;               // token availability
    sc_mutex mutex;                                   // Mutex to ensure atomic operations
    std::vector<std::unordered_set<int>> queue_entries; // Track entries in each queue to prevent duplicates
    int last_pe_owner;  // to track last owner for round-robin assignment

    //print log messages
    void log_info(std::string msg){
        SC_REPORT_INFO(LOG_NAME, msg.c_str());
    }

    void log_error(std::string msg){
        SC_REPORT_ERROR(LOG_NAME, msg.c_str());
    }

};

#endif // TOKEN_MANAGER_H

